# Copyright 2020 Okera Inc. All Rights Reserved.

import boto3
import botocore
import requests
from pathlib import Path
import os

from botocore.credentials import DeferredRefreshableCredentials, CredentialProvider

USERS_TOKEN_PATH = (os.getenv("HOME") + '/.okera')

class OkeraException(Exception):
    pass

class OkeraCredentialProvider(CredentialProvider):
    CANONICAL_NAME = "okera-aws-creds"

    def __init__(self, rest_api_url, user, token, token_source=config, verify_ssl=False):
        super().__init__()
        self._okera_rest_api_url = rest_api_url.rstrip('/')
        self._okera_token = _resolve_token(user, token, token_source)
        self._verify_ssl = verify_ssl

    def load(self):
        creds = DeferredRefreshableCredentials(refresh_using=self._refresh, method="sts-assume-role")
        return creds

    def _refresh(self):
        response = self._custom_aws_cred_refresh()
        credentials = {
            "access_key": response.get("key"),
            "secret_key": response.get("secret"),
            "expiry_time": response.get("expiry"),
            "token": None,
        }
        return credentials

    def _custom_aws_cred_refresh(self):
        api_url = "%s/%s" % (self._okera_rest_api_url, "api/v2/aws-tokens")
        headers = {"Authorization": "Bearer %s" % self._okera_token}
        res = requests.post(api_url, headers=headers, verify=self._verify_ssl)
        if res.status_code in (401, 403, 405):
            raise OkeraException("Error in authenticating credentials request: %s" % res.text)
        else:
            return res.json()

    def _read_token(token_file, user):
        try:
            with open(token_file, 'r') as file:
                data = file.read()
                return data
        except OSError as e:
            if (e.errno == errno.EACCES):
                raise OkeraException("No read access to %s's token file." % user, s)
            else:
                raise OkeraException("Error reading user's token file.", e)


    def _resolve_token(user, token, token_source):
        # token in home dir is priority, otherwise token_source
        home_token = Path(USERS_TOKEN_PATH)
        if home_token.is_file():
            return _read_token(home_token, user)
        elif ( "authserver" == token_source ):
            auth_uri=(_AUTH_SERVER_URI + "/" + user)
            res = requests.get(auth_uri)
            if res.status_code in (401, 403, 405):
                raise OkeraException("Error in retrieving user's JWT token: %s" % res.text)
            else:
                token_file = res.text
            data = _read_token(token_file, user)
            return data
        elif ( "config" == token_source ):
            return token

def okera_session(session, user, token, token_source, rest_uri, proxy_uri):
    bc_session = None
    if isinstance(session, botocore.session.Session):
        bc_session = session
    else:
        bc_session = session._session

    boto3_session = session

    # TODO: don't assume it's a cred resolver
    cred_chain = bc_session.get_component('credential_provider')
    okera_cred_provider = OkeraCredentialProvider(rest_uri, user, token, token_source)
    if cred_chain.providers:
        first_provider = cred_chain.providers[0].METHOD
        cred_chain.insert_before(first_provider, okera_cred_provider)
    else:
        cred_chain.providers.insert(0, okera_cred_provider)
    config = botocore.config.Config(
        proxies={'https': proxy_uri})

    orig_config = bc_session.get_default_client_config()
    if not orig_config:
        orig_config = botocore.config.Config()

    bc_session.set_default_client_config(orig_config.merge(config))

    return boto3_session


# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Some scenario tests for steward delegation
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

from okera import context
from okera.tests import pycerebro_test_common as common

from okera import _thrift_api

ATTR = "perf_test.attr1"
DB = "python_perf_test_db"
TBL1 = "tbl1"
TBL2 = "tbl2"
ROLE = "python_perf_test_role"
TEST_USER = "pythonperfuser"

def _create_wide_view(db, tbl, multiple, source_db, source_tbl):
    cols = []
    for idx in range(multiple):
        cols.extend([
            "uid AS uid%04d" % (idx),
            "dob AS dob%04d" % (idx),
            "gender AS gender%04d" % (idx),
            "ccn AS ccn%04d" % (idx),
        ])
    stmt = "CREATE VIEW %s.%s AS SELECT %s FROM %s.%s" % (
        db, tbl, ', '.join(cols), source_db, source_tbl)

    return stmt

class PerfTest(common.TestBase):
    def test_wide_table_with_attributes(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            print()
            ddls = [
                "DROP ATTRIBUTE IF EXISTS %s" % (ATTR),
                "CREATE ATTRIBUTE %s" % (ATTR),

                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "CREATE DATABASE %s" % (DB),

                """CREATE TABLE {db}.users (
                  uid STRING ATTRIBUTE {attr},
                  dob STRING ATTRIBUTE {attr},
                  gender STRING ATTRIBUTE {attr},
                  ccn STRING ATTRIBUTE {attr}
                )""".format(db=DB, attr=ATTR),

                _create_wide_view(DB, TBL1, 100, DB, 'users'),

                "DROP ROLE IF EXISTS %s" % (ROLE),
                "CREATE ROLE %s WITH GROUPS %s" % (ROLE, TEST_USER),
                "GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE IN (%s) TO ROLE %s" % (DB, ATTR, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            """
            ---------------------------- admin --------------------------
            Iterations 100
            Mean 68.4731936454773 ms
            50%: 67.74592399597168 ms
            90%: 70.44172286987305 ms
            95%: 73.26173782348633 ms
            99%: 100.9974479675293 ms
            99.5%: 100.9974479675293 ms
            99.9%: 100.9974479675293 ms
            """
            def list(tbl):
                def get():
                    datasets = conn.list_datasets(DB, name=tbl)
                    assert len(datasets) == 1
                return get

            common.measure_latency(100, list(TBL1))

            """
            ---------------------------- test user --------------------------
            Iterations 100
            Mean 484.99685287475586 ms
            50%: 477.36072540283203 ms
            90%: 510.06484031677246 ms
            95%: 526.5743732452393 ms
            99%: 825.0184059143066 ms
            99.5%: 825.0184059143066 ms
            99.9%: 825.0184059143066 ms
            """
            ctx.enable_token_auth(token_str=TEST_USER)
            common.measure_latency(100, list(TBL1))